#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "draw.h"
#include "caph.h"

#define	FB_ADDR(x,y)		(((color_t *)screen_pixels) + screen_w * (y) + (x))
#define FB_DRAW(x,y,col)	(*FB_ADDR((x),(y)) = (col))

static int
clip_code(point_t *s)
{
	int		code = 0;
	float		lim_w = (float) (screen_w - 1);
	float		lim_h = (float) (screen_h - 1);

	if (s->x < 0.0f)
		code |= 0x01;
	else if (s->x > lim_w)
		code |= 0x02;
	if (s->y < 0.0f)
		code |= 0x04;
	else if (s->y > lim_h)
		code |= 0x08;

	return code;
}

static int
clip_line(point_t *s, point_t *e)
{
	int		s_cc = clip_code(s);
	int		e_cc = clip_code(e);
	float		lim_w = (float) (screen_w - 1);
	float		lim_h = (float) (screen_h - 1);
	float		dx, dy;

	do {
		if (s_cc & e_cc)
			return -1;

		if (!(s_cc | e_cc))
			return 0;

		if (!s_cc) {

			int tmp = s_cc;
			s_cc = e_cc;
			e_cc = tmp;

			point_t *ptmp = s;
			s = e;
			e = ptmp;
		}

		dx = e->x - s->x;
		dy = e->y - s->y;

		if ((fabsf(dx) < 1.0f) && (s_cc & 0x03))
			return -1;

		if ((fabsf(dy) < 1.0f) && (s_cc & 0x0c))
			return -1;

		if (s_cc & 0x01) {
			s->y = e->y + ((0.0f - e->x) * dy) / dx;
			s->x = 0.0f;
			goto __cl_recalc;
		}

		if (s_cc & 0x02) {
			s->y = s->y + ((lim_w - s->x) * dy) / dx;
			s->x = lim_w;
			goto __cl_recalc;
		}

		if (s_cc & 0x04) {
			s->x = e->x + ((0.0f - e->y) * dx) / dy;
			s->y = 0.0f;
			goto __cl_recalc;
		}

		if (s_cc & 0x08) {
			s->x = s->x + ((lim_h - s->y) * dx) / dy;
			s->y = lim_h;
			goto __cl_recalc;
		}

__cl_recalc:
		s_cc = clip_code(s);
	}
	while (1);
}

static void
draw_int_line(int xs, int ys, int xe, int ye, color_t col)
{
	int	dx, dy;
	int	vx, vy;
	int	i = 0;

	if (xs < xe) {
		dx = xe - xs;
		vx = 1;
	}
	else {
		dx = xs - xe;
		vx = -1;
	}

	if (ys < ye) {
		dy = ye - ys;
		vy = 1;
	}
	else {
		dy = ys - ye;
		vy = -1;
	}

	if (dx < dy) {
		while (ys != ye) {
			FB_DRAW(xs, ys, col);

			ys += vy;
			i += dx;

			if (i >= dy) {
				i -= dy;
				xs += vx;
			}
		}
	}
	else if (dx > dy) {
		while (xs != xe) {
			FB_DRAW(xs, ys, col);

			xs += vx;
			i += dy;

			if (i >= dx) {
				i -= dx;
				ys += vy;
			}
		}
	}
	else {
		while (ys != ye) {
			FB_DRAW(xs, ys, col);

			xs += vx;
			ys += vy;
		}
	}

	FB_DRAW(xe, ye, col);
}

void draw_clear(color_t col)
{
	uint32_t *ptr;
	uint32_t *fin;
	uint32_t fill;

	fill = *(uint32_t*) &col;

	ptr = (uint32_t*) FB_ADDR(0, 0);
	fin = ptr + screen_w * screen_h;

	while (ptr != fin) {
		*ptr++ = fill;
		*ptr++ = fill;
		*ptr++ = fill;
		*ptr++ = fill;
	}
}

void draw_line(const point_t *s, const point_t *e, color_t col)
{
	point_t		ls, le;

	ls = *s;
	le = *e;

	if (clip_line(&ls, &le) < 0)
		return;

	draw_int_line(	(int) ls.x,
			(int) ls.y,
			(int) le.x,
			(int) le.y,
			col);
}

